@@ -2,6 +2,7 @@ import math
 from collections.abc import Callable
 from typing import Any
 
+from actions import physics_adapter as _pa
 from actions.base import Action as _Action
 
 
@@ -53,7 +54,7 @@ class MoveUntil(_Action):
         self.bounds = bounds  # (left, bottom, right, top)
         self.boundary_behavior = boundary_behavior
 
-        # New API: Velocity provider and boundary event callbacks
+        # Velocity provider and boundary event callbacks
         self.velocity_provider = velocity_provider
         self.on_boundary_enter = on_boundary_enter
         self.on_boundary_exit = on_boundary_exit
@@ -95,21 +96,12 @@ class MoveUntil(_Action):
             action="MoveUntil",
         )
 
-        # Try to extract duration from the condition function if it's from duration() helper
+        # Try to extract duration from explicit attribute if it's from duration() helper
         self._duration = None
-        try:
-            # Check if condition is from duration() helper by looking for closure
-            if (
-                hasattr(self.condition, "__closure__")
-                and self.condition.__closure__
-                and len(self.condition.__closure__) >= 1
-            ):
-                # Get the seconds value from the closure
-                seconds = self.condition.__closure__[0].cell_contents
-                if isinstance(seconds, (int, float)) and seconds > 0:
-                    self._duration = seconds
-        except (AttributeError, IndexError, TypeError):
-            pass
+        if hasattr(self.condition, "_duration_seconds"):
+            seconds = self.condition._duration_seconds
+            if isinstance(seconds, (int, float)) and seconds > 0:
+                self._duration = seconds
 
         self._elapsed = 0.0
 
@@ -190,8 +182,7 @@ class MoveUntil(_Action):
                     sprite.change_y = dy
             else:
                 # Normal behavior for other boundary types or no boundaries
-                sprite.change_x = dx
-                sprite.change_y = dy
+                _pa.set_velocity(sprite, (dx, dy))
 
         self.for_each_sprite(set_velocity)
 
@@ -245,10 +236,7 @@ class MoveUntil(_Action):
                             # Trigger boundary enter event if not already at boundary
                             if state["x"] != "right":
                                 if self.on_boundary_enter:
-                                    try:
-                                        self.on_boundary_enter(sprite, "x", "right")
-                                    except Exception:
-                                        pass
+                                    self._safe_call(self.on_boundary_enter, sprite, "x", "right")
                                 state["x"] = "right"
                         elif dx < 0 and sprite.center_x + dx < left:
                             sprite.change_x = 0
@@ -256,10 +244,7 @@ class MoveUntil(_Action):
                             # Trigger boundary enter event if not already at boundary
                             if state["x"] != "left":
                                 if self.on_boundary_enter:
-                                    try:
-                                        self.on_boundary_enter(sprite, "x", "left")
-                                    except Exception:
-                                        pass
+                                    self._safe_call(self.on_boundary_enter, sprite, "x", "left")
                                 state["x"] = "left"
                         else:
                             sprite.change_x = dx
@@ -277,10 +262,7 @@ class MoveUntil(_Action):
                             # Trigger boundary enter event if not already at boundary
                             if state["y"] != "top":
                                 if self.on_boundary_enter:
-                                    try:
-                                        self.on_boundary_enter(sprite, "y", "top")
-                                    except Exception:
-                                        pass
+                                    self._safe_call(self.on_boundary_enter, sprite, "y", "top")
                                 state["y"] = "top"
                         elif dy < 0 and sprite.center_y + dy < bottom:
                             sprite.change_y = 0
@@ -288,10 +270,7 @@ class MoveUntil(_Action):
                             # Trigger boundary enter event if not already at boundary
                             if state["y"] != "bottom":
                                 if self.on_boundary_enter:
-                                    try:
-                                        self.on_boundary_enter(sprite, "y", "bottom")
-                                    except Exception:
-                                        pass
+                                    self._safe_call(self.on_boundary_enter, sprite, "y", "bottom")
                                 state["y"] = "bottom"
                         else:
                             sprite.change_y = dy
@@ -603,6 +582,10 @@ class FollowPathUntil(_Action):
     The action supports automatic sprite rotation to face the movement direction, with
     calibration offset for sprites that aren't naturally drawn pointing to the right.
 
+    Optional physics integration: When use_physics=True and a PymunkPhysicsEngine is
+    available, the action uses steering impulses to follow the path, allowing natural
+    interaction with other physics forces and collisions.
+
     Args:
         control_points: List of (x, y) points defining the Bezier curve (minimum 2 points)
         velocity: Speed in pixels per second along the curve
@@ -616,6 +599,11 @@ class FollowPathUntil(_Action):
             - -90.0: Sprite artwork points up
             - 180.0: Sprite artwork points left
             - 90.0: Sprite artwork points down
+        use_physics: When True, uses physics steering with impulses instead of kinematic
+            movement. Requires a PymunkPhysicsEngine. Default: False.
+        steering_gain: Tunable gain parameter for physics steering responsiveness.
+            Higher values = more responsive but may overshoot. Lower values = smoother
+            but may lag behind path. Default: 5.0. Only used when use_physics=True.
 
     Examples:
         # Basic path following without rotation
@@ -639,6 +627,12 @@ class FollowPathUntil(_Action):
             bezier_points, velocity=200, condition=lambda: sprite.center_x > 400,
             rotate_with_path=True
         )
+
+        # Physics-based path following with steering
+        action = FollowPathUntil(
+            [(100, 100), (300, 200), (500, 100)], velocity=150, condition=infinite,
+            use_physics=True, steering_gain=5.0, rotate_with_path=True
+        )
     """
 
     def __init__(
@@ -649,6 +643,8 @@ class FollowPathUntil(_Action):
         on_stop: Callable[[Any], None] | Callable[[], None] | None = None,
         rotate_with_path: bool = False,
         rotation_offset: float = 0.0,
+        use_physics: bool = False,
+        steering_gain: float = 5.0,
     ):
         super().__init__(condition, on_stop)
         if len(control_points) < 2:
@@ -659,6 +655,8 @@ class FollowPathUntil(_Action):
         self.current_velocity = velocity  # Current velocity (can be scaled)
         self.rotate_with_path = rotate_with_path  # Enable automatic sprite rotation
         self.rotation_offset = rotation_offset  # Degrees to offset for sprite artwork orientation
+        self.use_physics = use_physics  # Enable physics-based steering
+        self.steering_gain = steering_gain  # Steering force multiplier for physics mode
 
         # Track last applied movement angle to smooth out rotations when
         # the incremental movement vector becomes (nearly) zero – e.g. when
@@ -735,7 +733,18 @@ class FollowPathUntil(_Action):
         # Calculate new position on curve
         current_point = self._bezier_point(self._curve_progress)
 
-        # Apply relative movement to sprite(s)
+        # Check if physics engine is available for steering mode
+        engine = None
+        if self.use_physics and self.target is not None:
+            # Check if any sprite in target has physics
+            def check_engine(sprite):
+                nonlocal engine
+                if engine is None:
+                    engine = _pa.detect_engine(sprite)
+
+            self.for_each_sprite(check_engine)
+
+        # Apply movement using physics steering or kinematic mode
         if self._last_position:
             dx = current_point[0] - self._last_position[0]
             dy = current_point[1] - self._last_position[1]
@@ -755,15 +764,58 @@ class FollowPathUntil(_Action):
                     # Re-use the previous angle to maintain continuity
                     movement_angle = self._prev_movement_angle
 
-            def apply_movement(sprite):
-                # Move sprite along the path
-                sprite.center_x += dx
-                sprite.center_y += dy
-                # Rotate sprite to face movement direction if enabled
-                if movement_angle is not None:
-                    sprite.angle = movement_angle
+            if engine is not None:
+                # Physics steering mode: apply impulses to steer toward path
+                def apply_physics_steering(sprite):
+                    # Get current sprite position and desired target
+                    current_pos = (sprite.center_x, sprite.center_y)
 
-            self.for_each_sprite(apply_movement)
+                    # Compute desired velocity toward next point on curve
+                    direction_length = math.sqrt(dx * dx + dy * dy)
+                    if direction_length > 1e-6:
+                        desired_vx = (dx / direction_length) * self.current_velocity
+                        desired_vy = (dy / direction_length) * self.current_velocity
+                    else:
+                        desired_vx = desired_vy = 0.0
+
+                    # Get current velocity from physics engine
+                    current_vx, current_vy = _pa.get_velocity(sprite)
+
+                    # Compute steering impulse (desired - current) * gain
+                    # Note: In a real physics engine, we'd multiply by mass here,
+                    # but for simplicity we use a tunable gain parameter
+                    steering_x = (desired_vx - current_vx) * self.steering_gain * delta_time
+                    steering_y = (desired_vy - current_vy) * self.steering_gain * delta_time
+
+                    # Apply steering impulse
+                    _pa.apply_impulse(sprite, (steering_x, steering_y))
+
+                    # Handle rotation via physics engine if enabled
+                    if movement_angle is not None:
+                        # For physics mode, we set angular velocity to reach target angle
+                        # Calculate difference and apply proportional angular velocity
+                        angle_diff = movement_angle - sprite.angle
+                        # Normalize to [-180, 180]
+                        while angle_diff > 180:
+                            angle_diff -= 360
+                        while angle_diff < -180:
+                            angle_diff += 360
+                        # Apply angular velocity proportional to angle difference
+                        angular_vel = angle_diff * 5.0  # Simple proportional controller
+                        _pa.set_angular_velocity(sprite, angular_vel)
+
+                self.for_each_sprite(apply_physics_steering)
+            else:
+                # Kinematic mode: directly set position (original behavior)
+                def apply_movement(sprite):
+                    # Move sprite along the path
+                    sprite.center_x += dx
+                    sprite.center_y += dy
+                    # Rotate sprite to face movement direction if enabled
+                    if movement_angle is not None:
+                        sprite.angle = movement_angle
+
+                self.for_each_sprite(apply_movement)
 
         self._last_position = current_point
 
@@ -801,6 +853,8 @@ class FollowPathUntil(_Action):
             self.on_stop,
             self.rotate_with_path,
             self.rotation_offset,
+            self.use_physics,
+            self.steering_gain,
         )
 
 
@@ -838,7 +892,7 @@ class RotateUntil(_Action):
         """Apply angular velocity to all sprites."""
 
         def set_angular_velocity(sprite):
-            sprite.change_angle = self.current_angular_velocity
+            _pa.set_angular_velocity(sprite, self.current_angular_velocity)
 
         self.for_each_sprite(set_angular_velocity)
 
@@ -1116,21 +1170,12 @@ class DelayUntil(_Action):
 
     def apply_effect(self) -> None:
         """Initialize delay timing."""
-        # Try to extract duration from the condition function if it's from duration() helper
+        # Try to extract duration from explicit attribute if it's from duration() helper
         self._duration = None
-        try:
-            # Check if condition is from duration() helper by looking for closure
-            if (
-                hasattr(self.condition, "__closure__")
-                and self.condition.__closure__
-                and len(self.condition.__closure__) >= 1
-            ):
-                # Get the seconds value from the closure
-                seconds = self.condition.__closure__[0].cell_contents
-                if isinstance(seconds, (int, float)) and seconds > 0:
-                    self._duration = seconds
-        except (AttributeError, IndexError, TypeError):
-            pass
+        if hasattr(self.condition, "_duration_seconds"):
+            seconds = self.condition._duration_seconds
+            if isinstance(seconds, (int, float)) and seconds > 0:
+                self._duration = seconds
 
         self._elapsed = 0.0
 
@@ -1259,16 +1304,12 @@ class TweenUntil(_Action):
         self._factor = factor
 
     def apply_effect(self):
-        # Extract duration (explicit or closure) FIRST
+        # Extract duration from explicit attribute
         duration_val = 1.0
-        try:
-            # EAFP: Try to get duration from the closure of the `duration` helper.
-            # This is more Pythonic and robust than checking function name with hasattr.
-            if self.condition.__name__ == "condition":
-                duration_val = self.condition.__closure__[0].cell_contents
-        except (AttributeError, IndexError, TypeError):
-            # This is expected if condition is not from `duration()` or has no closure.
-            pass
+        if hasattr(self.condition, "_duration_seconds"):
+            seconds = self.condition._duration_seconds
+            if isinstance(seconds, (int, float)):
+                duration_val = seconds
 
         # An explicitly set duration should override the one from the condition.
         if self._duration is not None:
@@ -1278,13 +1319,10 @@ class TweenUntil(_Action):
         if self._duration < 0:
             raise ValueError("Duration must be non-negative")
 
-        # Define a helper to set the initial value, using EAFP for property validation.
+        # Define a helper to set the initial value.
         def set_initial_value(sprite):
             """Set the initial value of the property on a single sprite."""
-            try:
-                setattr(sprite, self.property_name, self.start_value)
-            except AttributeError:
-                raise AttributeError(f"Target sprite does not have property '{self.property_name}'")
+            setattr(sprite, self.property_name, self.start_value)
 
         if self._duration == 0:
             # If duration is zero, immediately set to the end value.
@@ -1395,6 +1433,7 @@ class CallbackUntil(_Action):
         _debug_log(f"__init__: id={id(self)}, callback={callback}, seconds_between_calls={seconds_between_calls}")
 
         # If the condition is a duration() helper, replace it with a simulation-time condition
+        # for more accurate timing. If optimization fails, fall back to original condition.
         try:
             if hasattr(condition, "_is_duration_condition") and condition._is_duration_condition:
                 seconds = getattr(condition, "_duration_seconds", None)
@@ -1408,8 +1447,9 @@ class CallbackUntil(_Action):
                     _sim_condition._is_duration_condition = True
                     _sim_condition._duration_seconds = self._duration
                     self.condition = _sim_condition
-        except Exception:
-            pass
+        except (AttributeError, TypeError) as e:
+            # Duration optimization failed, fall back to original condition
+            _debug_log(f"__init__: id={id(self)}, duration optimization failed: {e}, using original condition")
 
     def set_factor(self, factor: float) -> None:
         """Scale the callback interval by the given factor.
@@ -1499,34 +1539,12 @@ class CallbackUntil(_Action):
         self._elapsed_since_call = 0.0
         self.current_seconds_between_calls = self.target_seconds_between_calls
         self._next_fire_time = None
-        # Try to extract duration from the condition function if it's from duration() helper
+        # Try to extract duration from explicit attribute if it's from duration() helper
         self._duration = None
-        # Prefer attribute set by duration() helper if available
-        try:
-            if hasattr(self.condition, "_duration_seconds"):
-                seconds = self.condition._duration_seconds
-                if isinstance(seconds, (int, float)) and seconds >= 0:
-                    self._duration = seconds
-        except Exception:
-            pass
-        if self._duration is None:
-            try:
-                if (
-                    hasattr(self.condition, "__closure__")
-                    and self.condition.__closure__
-                    and len(self.condition.__closure__) >= 1
-                ):
-                    seconds = self.condition.__closure__[0].cell_contents
-                    if isinstance(seconds, (int, float)) and seconds >= 0:
-                        self._duration = seconds
-            except (AttributeError, IndexError, TypeError):
-                pass
-        # Fallback to helper if still None
-        if self._duration is None:
-            try:
-                self._duration = _extract_duration_seconds(self.condition)
-            except Exception:
-                self._duration = None
+        if hasattr(self.condition, "_duration_seconds"):
+            seconds = self.condition._duration_seconds
+            if isinstance(seconds, (int, float)) and seconds >= 0:
+                self._duration = seconds
 
     def _call_callback_with_fallback(self) -> None:
         """Call the callback, trying both with and without target parameter."""
@@ -1684,24 +1702,15 @@ class ParametricMotionUntil(_Action):
         self.for_each_sprite(capture_origin)
 
         if self._duration is None:
-            # Try to extract duration from the condition function if it's from duration() helper
-            try:
-                # Check if condition is from duration() helper by looking for closure
-                if (
-                    hasattr(self.condition, "__closure__")
-                    and self.condition.__closure__
-                    and len(self.condition.__closure__) >= 1
-                ):
-                    # Get the seconds value from the closure
-                    seconds = self.condition.__closure__[0].cell_contents
-                    if isinstance(seconds, (int, float)) and seconds > 0:
-                        self._duration = seconds
-            except (AttributeError, IndexError, TypeError):
-                pass
+            # Try to extract duration from explicit attribute if it's from duration() helper
+            if hasattr(self.condition, "_duration_seconds"):
+                seconds = self.condition._duration_seconds
+                if isinstance(seconds, (int, float)) and seconds > 0:
+                    self._duration = seconds
 
-            # Fallback to the helper function if the above didn't work
+            # If still None, default to 0.0
             if self._duration is None or self._duration == 0:
-                self._duration = _extract_duration_seconds(self.condition) or 0.0
+                self._duration = 0.0
 
         # Do not pre-position sprites; offsets are relative to captured origins
         self._prev_offset = self._offset_fn(0.0)
@@ -1810,11 +1819,11 @@ def _apply_offset(sprite, dx: float, dy: float, origins: dict[int, tuple[float,
 
 
 def _extract_duration_seconds(cond: _Callable[[], _Any]) -> float | None:
-    try:
-        if cond.__name__ == "condition":
-            return cond.__closure__[0].cell_contents  # type: ignore[index]
-    except Exception:
-        pass
+    """Extract duration from explicit attribute if available."""
+    if hasattr(cond, "_duration_seconds"):
+        seconds = cond._duration_seconds
+        if isinstance(seconds, (int, float)) and seconds >= 0:
+            return seconds
     return None
 
 
@@ -1857,27 +1866,22 @@ class CycleTexturesUntil(_Action):
         self._elapsed = 0.0
         self._duration: float | None = None
 
-        # Try to extract duration from duration() helper functions
-        try:
-            if hasattr(condition, "__name__") and condition.__name__ == "condition":
-                if hasattr(condition, "__closure__") and condition.__closure__:
-                    seconds = condition.__closure__[0].cell_contents
-                    if isinstance(seconds, (int, float)) and seconds >= 0:
-                        self._duration = float(seconds)
+        # Try to extract duration from explicit attribute if it's from duration() helper
+        if hasattr(condition, "_duration_seconds"):
+            seconds = condition._duration_seconds
+            if isinstance(seconds, (int, float)) and seconds >= 0:
+                self._duration = float(seconds)
 
-                        def _sim_condition() -> bool:
-                            return self._elapsed >= (self._duration or 0.0) - 1e-9
+                def _sim_condition() -> bool:
+                    return self._elapsed >= (self._duration or 0.0) - 1e-9
 
-                        # Preserve attributes so cloning and tools can still introspect
-                        _sim_condition._is_duration_condition = True
-                        _sim_condition._original_condition = condition
-                        _sim_condition._duration_seconds = self._duration
+                # Preserve attributes so cloning and tools can still introspect
+                _sim_condition._is_duration_condition = True
+                _sim_condition._original_condition = condition
+                _sim_condition._duration_seconds = self._duration
 
-                        # Replace the condition with simulation-time version
-                        self.condition = _sim_condition
-        except Exception:
-            # If duration extraction fails, use original condition
-            pass
+                # Replace the condition with simulation-time version
+                self.condition = _sim_condition
 
     def apply_effect(self) -> None:
         """Initialize textures on the target sprite(s)."""
@@ -1886,14 +1890,10 @@ class CycleTexturesUntil(_Action):
 
         # Try duration extraction again in case condition wasn't extractable during __init__
         if self._duration is None:
-            try:
-                if hasattr(self.condition, "__name__") and self.condition.__name__ == "condition":
-                    if hasattr(self.condition, "__closure__") and self.condition.__closure__:
-                        seconds = self.condition.__closure__[0].cell_contents
-                        if isinstance(seconds, (int, float)) and seconds >= 0:
-                            self._duration = float(seconds)
-            except Exception:
-                pass
+            if hasattr(self.condition, "_duration_seconds"):
+                seconds = self.condition._duration_seconds
+                if isinstance(seconds, (int, float)) and seconds >= 0:
+                    self._duration = float(seconds)
 
         def set_initial_texture(sprite):
             sprite.textures = self._textures
