@@ -120,7 +120,7 @@ class Starfield:
             condition=infinite,
             bounds=bounds,
             boundary_behavior="wrap",
-            on_boundary=self.on_stars_wrap,
+            on_boundary_exit=self.on_stars_wrap,
         )
 
     def update(self):
